#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""filt.py
@author: Louis Richard
"""

import xarray as xr
import numpy as np

from scipy import signal


# noinspection PyTupleAssignmentBalance
def filt(inp, f_min=0., f_max=1., order=-1):
    """Filters input quantity.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the variable to filter.

    f_min : float
        Lower limit of the frequency range.

    f_max : float
        Upper limit of the frequency range.

    order : int
        Order of the elliptic filter.

    Returns
    -------
    out : xarray.DataArray
        Time series of the filtered signal.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2017-07-18T13:03:34.000", "2017-07-18T13:07:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic and electric fields

    >>> b_xyz = mms.get_data("B_gse_fgm_brst_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_brst_l2", tint, mms_id)

    Convert E to field aligned coordinates

    >>> e_xyzfac = pyrf.convert_fac(e_xyz, b_xyz, [1,0,0])

    Bandpass filter E waveform

    >>> e_xyzfac_hf = pyrf.filt(e_xyzfac, 4, 0, 3)
    >>> e_xyzfac_lf = pyrf.filt(e_xyzfac, 0, 4, 3)

    """

    f_samp = 1 / (np.median(np.diff(inp.time)).astype(int) * 1e-9)

    # Data of the input
    inp_data = inp.data

    f_min, f_max = [f_min / (f_samp / 2), f_max / (f_samp / 2)]

    f_max = np.min([f_max, 1.])

    # Parameters of the elliptic filter. fact defines the width between
    # stopband and passband
    # r_pass, r_stop, fact = [0.5, 60, 1.1]

    if f_min == 0:
        num1, den1, num2, den2 = [None] * 4

        if order == -1:
            order, f_max = signal.ellipord(f_max,
                                           np.min([f_max * 1.1, 0.9999]),
                                           .5, 60)

        num, den = signal.ellip(order, .5, 60, f_max, btype="lowpass")
    elif f_max == 0:
        num1, den1, num2, den2 = [None] * 4

        if order == -1:
            order, f_min = signal.ellipord(f_min,
                                           np.min([f_min * 1.1, 0.9999]),
                                           .5, 60)

        num, den = signal.ellip(order, .5, 60, f_min, btype="highpass")
    else:
        num, den = [None] * 2

        if order == -1:
            order, f_max = signal.ellipord(f_max,
                                           np.min([f_max * 1.3, 0.9999]),
                                           .5, 60)

        num1, den1 = signal.ellip(order, .5, 60, f_max)

        if order == -1:
            order, f_min = signal.ellipord(f_min, f_min * .75, .5, 60)

        num2, den2 = signal.ellip(order, .5, 60, f_min)

    if len(inp_data.shape) == 1:
        inp_data = inp_data[:, np.newaxis]

    out_data = np.zeros(inp_data.shape)

    for i_col in range(inp_data.shape[1]):
        if f_min != 0 and f_max != 0:
            out_data[:, i_col] = signal.filtfilt(num1, den1,
                                                 inp_data[:, i_col])
            out_data[:, i_col] = signal.filtfilt(num2, den2,
                                                 out_data[:, i_col])
        else:
            out_data[:, i_col] = signal.filtfilt(num, den, inp_data[:, i_col])

    if inp_data.shape[1] == 1:
        out_data = out_data[:, 0]

    out = xr.DataArray(out_data, coords=inp.coords, dims=inp.dims,
                       attrs=inp.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""wave_fft.py
@author: Louis Richard
"""

import numpy as np

from scipy import signal


def wave_fft(inp, window, frame_overlap: float = 10.,
             frame_length: float = 20., f_sampling: float = None):
    """Short-Time Fourier Transform.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the one dimension data.

    window : str
        Window function such as rectwin, hamming (default).

    frame_overlap : float, optional
        Length of each frame overlaps in second.

    frame_length : float, optional
        Length of each frame in second.

    f_sampling : float, optional
        Sampling frequency.

    Returns
    -------
    spectrogram : ndarray
        Spectrogram of x.

    time : ndarray
        Value corresponds to the center of each frame (x-axis) in sec.

    frequencies : ndarray
        Vector of frequencies (y-axis) in Hz.

    """

    if f_sampling is None:
        delta_t = np.median(np.diff(inp.time.data).astype(float)) * 1e-9
        f_sampling = 1 / delta_t

    # convert ms to points
    n_per_seg = np.round(frame_length * f_sampling).astype(int)
    n_overlap = np.round(frame_overlap * f_sampling).astype(int)

    options = dict(fs=f_sampling, window=window, nperseg=n_per_seg,
                   noverlap=n_overlap, mode='complex')
    frequencies, time, spectrogram = signal.spectrogram(inp, **options)

    return frequencies, time, spectrogram



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.dates as mdates

from matplotlib import colors
from pandas.plotting import register_matplotlib_converters
register_matplotlib_converters()

plt.style.use("seaborn-whitegrid")
locator = mdates.AutoDateLocator(minticks=3, maxticks=7)
formatter = mdates.ConciseDateFormatter(locator)
sns.set_context("paper")
# plt.rc('lines', linewidth=1)


def plot_spectr(axis=None, inp=None, yscale="", cscale="", clim=None, cmap="",
                cbar=True, **kwargs):
    """Plot a spectrogram using pcolormesh.

    Parameters
    ----------
    axis : axes
        Target axis to plot. If None create a new figure.

    inp : DataArray
        Input 2D data to plot.

    yscale : str
        Y-axis flag. Default is "" (linear).

    cscale : str
        C-axis flag. Default is "" (linear).

    clim : list
        C-axis bounds. Default is None (autolim).

    cmap : str
        Colormap. Default is jet.

    cbar : bool
        Flag for colorbar. Set to False to hide.

    Returns
    -------
    fig : figure
        to fill.

    axs : axes
        to fill.

    caxs : caxes
        Only if cbar is True.

    """

    if axis is None:
        fig, axis = plt.subplots(1)
    else:
        fig = plt.gcf()

    if not cmap:
        cmap = "jet"

    if cscale == "log":
        if clim is not None and isinstance(clim, list):
            options = dict(norm=colors.LogNorm(vmin=clim[0], vmax=clim[1]),
                           cmap=cmap)
        else:
            options = dict(norm=colors.LogNorm(), cmap=cmap)
    else:
        if clim is not None and isinstance(clim, list):
            options = dict(cmap=cmap, vmin=clim[0], vmax=clim[1])
        else:
            options = dict(cmap=cmap, vmin=None, vmax=None)

    x_data, y_data = [inp.coords[inp.dims[0]], inp.coords[inp.dims[1]]]

    image = axis.pcolormesh(x_data, y_data, inp.data.T, rasterized=True,
                            **options)

    if yscale == "log":
        axis.set_yscale("log")

    if cbar:
        if kwargs.get("pad"):
            pad = kwargs["pad"]
        else:
            pad = 0.01

        pos = axis.get_position()
        cax = fig.add_axes([pos.x0+pos.width+pad, pos.y0, 0.01, pos.height])
        fig.colorbar(mappable=image, cax=cax, ax=axis)

        out = (axis, cax)
    else:
        out = axis

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""__init__.py
@author: Louis Richard
"""

# @Louis Richard
from .igrf import igrf
from .magnetopause_normal import magnetopause_normal



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""poynting_flux.py
@author: Louis Richard
"""

import numpy as np

from astropy.time import Time

from .calc_fs import calc_fs
from .cross import cross
from .dot import dot
from .normalize import normalize
from .resample import resample
from .time_clip import time_clip


def poynting_flux(e_xyz, b_xyz, b_hat):
    """
    Estimates Poynting flux at electric field sampling as

    .. math::

        \\mathbf{S} = \\frac{\\mathbf{E}\\times\\mathbf{B}}{\\mu_0}

    if `b0` is given project the Poynting flux along `b0`


    Parameters
    ----------
    e_xyz : xarray.DataArray
        Time series of the electric field.

    b_xyz : xarray.DataArray
        Time series of the magnetic field.

    b_hat : xarray.DataArray, optional
        Time series of the direction to project the Pointing flux.

    Returns
    -------
    s : xarray.DataArray
        Time series of the Pointing flux.

    s_z : xarray.DataArray
        Time series of the projection of the Pointing flux
        (only if b0).

    int_s : xarray.DataArray
        Time series of the time integral of the Pointing flux
        (if b0 integral along b0).

    """

    # check which Poynting flux to calculate
    flag_s_z, flag_int_s_z, flag_int_s = [False, False, False]

    if b_hat is None:
        flag_int_s = True
    else:
        flag_s_z, flag_int_s_z = [True, True]

    # interval where both E & B exist
    tint = [Time(max([min(e_xyz.time.data), min(b_xyz.time.data)]),
                 format="datetime64").iso,
            Time(min([max(e_xyz.time.data), max(b_xyz.time.data)]),
                 format="datetime64").iso]

    e_xyz, b_xyz = [time_clip(e_xyz, tint), time_clip(b_xyz, tint)]

    if len(e_xyz) < len(b_xyz):
        e_xyz = resample(e_xyz, b_xyz)
        f_spl = calc_fs(b_xyz)
    elif len(e_xyz) > len(b_xyz):
        b_xyz = resample(b_xyz, e_xyz)
        f_spl = calc_fs(e_xyz)
    else:
        f_spl = calc_fs(b_xyz)

    # Calculate Poynting flux
    s_xyz = cross(e_xyz, b_xyz) / (4 * np.pi / 1e7) * 1e-9

    if flag_s_z:
        b_m = resample(b_hat, e_xyz)
        s_z = dot(normalize(b_m), s_xyz)
    else:
        s_z = None

    # time integral of Poynting flux along ambient magnetic field
    res = None

    if flag_int_s_z:
        s_z[np.isnan(s_z.data)] = 0  # set to zero points where Sz=NaN

        int_s_z = np.cumsum(s_z) / f_spl

        res = (s_xyz, s_z, int_s_z)

    if flag_int_s:  # time integral of all Poynting flux components
        # set to zero points where Sz=NaN
        s_xyz[np.isnan(s_xyz[:, 2].data)] = 0

        int_s = np.cumsum(s_xyz) / f_spl

        res = (s_xyz, int_s)

    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""solid_angle.py
@author: Louis Richard
"""

import numpy as np


def solid_angle(inp0, inp1, inp2):
    """Calculates the solid angle of three vectors making up a triangle
    in a unit sphere with the sign taken into account.

    Parameters
    ----------
    inp0 : ndarray
        First vector.

    inp1 : ndarray
        Second vector.

    inp2 : ndarray
        Third vector.

    Returns
    -------
    angle : float
        Solid angle.

    """

    # Calculate the smaller angles between the vectors around origin
    acos_12 = np.arccos(np.sum(inp2 * inp1))
    acos_02 = np.arccos(np.sum(inp0 * inp2))
    acos_01 = np.arccos(np.sum(inp1 * inp0))

    # Calculate the angles in the spherical triangle (Law of Cosines)
    alpha = np.arccos((np.cos(acos_12) - np.cos(acos_02) * np.cos(acos_01))
                  / (np.sin(acos_02) * np.sin(acos_01)))
    beta = np.arccos((np.cos(acos_02) - np.cos(acos_12) * np.cos(acos_01))
                  / (np.sin(acos_12) * np.sin(acos_01)))
    gamma = np.arccos((np.cos(acos_01) - np.cos(acos_02) * np.cos(acos_12))
                  / (np.sin(acos_02) * np.sin(acos_12)))

    # Calculates the Surface area on the unit sphere (solid angle)
    angle = (alpha + beta + gamma - np.pi)
    # Calculate the sign of the area
    var = np.cross(inp2, inp1)
    div = np.sum(var * inp0)
    sgn = np.sign(div)

    # Solid angle with sign taken into account
    angle = sgn * angle

    return angle



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

import warnings
import tqdm
import numpy as np
import xarray as xr

from ..pyrf import ts_scalar, time_clip, resample, normalize


def get_pitch_angle_dist(vdf, b_xyz, tint, **kwargs):
    """
    Computes the pitch angle distributions from l1b brst particle data.

    Parameters
    ----------
    vdf : xarray.Dataset
        to fill

    b_xyz : xarray.DataArray
        to fill

    tint : list of str, optional
        Time interval for closeup.

    **kwargs : dict
        Hash table of keyword arguments with :
            * angles : int or float or list of ndarray
                User defined angles

            * meanorsum : str
                Method :
                    * mean : to fill
                    * sum : to fill
                    * sum_weighted : to fill


    Returns
    -------
    pad : xarray.DataArray
        Particle pitch angle distribution


    Examples
    --------
    >>> from pyrfu import mms

    Define time intervals

    >>> tint_long = ["2017-07-24T12:48:34.000", "2017-07-24T12:58:20.000"]
    >>> tint_zoom = ["2017-07-24T12:49:18.000", "2017-07-24T12:49:30.000"]

    Load ions velocity distribution for MMS1

    >>> vdf_i = mms.get_data("pdi_fpi_brst_l2", tint_long, 1)

    Load magnetic field in the spacecraft coordinates system.

    >>> b_dmpa = mms.get_data("b_dmpa_fgm_brst_l2", tint_long, 1)

    Compute pitch angle distribution

    >>> options = dict(angles=24)
    >>> pad_i = mms.get_pitch_angle_dist(vdf, b_dmpa, tint_zoom, **options)

    """

    # Default pitch angles. 15 degree angle widths
    angles_v = np.linspace(15, 180, int(180 / 15))
    d_angles = np.median(np.diff(angles_v)) * np.ones(len(angles_v))

    # Method
    mean_or_sum = "mean"

    if "angles" in kwargs:
        if isinstance(kwargs["angles"], (int, float)):
            n_angles = int(kwargs["angles"])  # Make sure input is integer
            d_angles = 180 / n_angles
            angles_v = np.linspace(d_angles, 180, n_angles)
            d_angles = d_angles * np.ones(n_angles)
            print("notice : User defined number of pitch angles.")

        elif isinstance(kwargs["angles"], (list, np.ndarray)):
            angles_v = kwargs["angles"]
            d_angles = np.diff(angles_v)
            angles_v = angles_v[1:]
            print("notice : User defined pitch angle limits.")

        else:
            raise ValueError("angles parameter not understood.")

    if "meanorsum" in kwargs:
        if isinstance(kwargs["meanorsum"], str) and kwargs["meanorsum"] in ["mean", "sum"]:
            mean_or_sum = kwargs["meanorsum"]
        else:
            raise ValueError("meanorsum parameter not understood.")

    pitch_angles = angles_v - d_angles / 2
    n_angles = len(angles_v)

    time = vdf.time.data

    vdf0 = vdf.data.copy()

    if vdf.phi.data.ndim == 1:
        phi = np.tile(vdf.phi.data, (len(time), 1))
        phi = xr.DataArray(phi, coords=[time, np.arange(len(phi))], dims=["time", "idx1"])
    else:
        phi = vdf.phi

    theta = vdf.theta

    if "esteptable" in vdf.attrs.keys():
        step_table = ts_scalar(time, vdf.attrs["esteptable"])
    else:
        step_table = ts_scalar(time, np.zeros(len(time)))

    if "energy0" in vdf.attrs.keys() and "energy1" in vdf.attrs.keys():
        energy0, _ = [vdf.attrs[f"energy{i}"] for i in range(2)]
    else:
        energy0, _ = vdf.energy.data[:2, :]

    if tint is not None:
        b_xyz = time_clip(b_xyz, tint)
        vdf0, phi, step_table = [time_clip(dat, tint) for dat in [vdf0, phi, step_table]]

    time = vdf0.time.data

    # Check size of energy
    n_en, n_phi, n_theta = [len(energy0), len(phi.data[0, :]), len(theta)]

    b_xyz = resample(b_xyz, vdf0)
    b_vec = normalize(b_xyz)

    b_vec_x = np.transpose(np.tile(b_vec.data[:, 0], [n_en, n_phi, n_theta, 1]), [3, 0, 1, 2])
    b_vec_y = np.transpose(np.tile(b_vec.data[:, 1], [n_en, n_phi, n_theta, 1]), [3, 0, 1, 2])
    b_vec_z = np.transpose(np.tile(b_vec.data[:, 2], [n_en, n_phi, n_theta, 1]), [3, 0, 1, 2])

    x_vec = np.zeros((len(time), n_phi, n_theta))
    y_vec = np.zeros((len(time), n_phi, n_theta))
    z_vec = np.zeros((len(time), n_phi, n_theta))

    for i in range(len(time)):
        x_vec[i, ...] = np.dot(-np.cos(np.deg2rad(phi.data[i, None])).T,
                               np.sin(np.deg2rad(theta.data[:, None])).T)
        y_vec[i, ...] = np.dot(-np.sin(np.deg2rad(phi.data[i, None])).T,
                               np.sin(np.deg2rad(theta.data[:, None])).T)
        z_vec[i, ...] = np.dot(-np.ones((n_phi, 1)), np.cos(np.deg2rad(theta.data[:, None])).T)

    if tint is not None:
        energy = time_clip(vdf.energy, tint).data
    else:
        energy = vdf.energy.data

    temp = [np.tile(mat, [n_en, 1, 1, 1]) for mat in [x_vec, y_vec, z_vec]]
    x_mat, y_mat, z_mat = [np.squeeze(np.transpose(mat, [1, 0, 2, 3])) for mat in temp]

    theta_b = np.arccos(x_mat * np.squeeze(b_vec_x) + y_mat * np.squeeze(b_vec_y)
                        + z_mat * np.squeeze(b_vec_z))
    theta_b = np.rad2deg(theta_b)

    dists = [vdf0.data.copy() for _ in range(n_angles)]

    pad_arr = [None] * n_angles

    for i in tqdm.tqdm(np.arange(n_angles)):
        dists[i][theta_b < angles_v[i] - d_angles[i]] = np.nan
        dists[i][theta_b > angles_v[i]] = np.nan

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            if mean_or_sum == "mean":
                pad_arr[i] = np.squeeze(np.nanmean(np.nanmean(dists[i], axis=3), axis=2))
            elif mean_or_sum == "sum":
                pad_arr[i] = np.squeeze(np.nansum(np.nansum(dists[i], axis=3), axis=2))
            else:
                raise ValueError("Invalid method")

    pad_arr = np.transpose(np.stack(pad_arr), [1, 0, 2])

    energy = np.mean(energy[:2, :], axis=0)

    pad = xr.Dataset({"data": (["time", "idx0", "idx1"], np.transpose(pad_arr, [0, 2, 1])),
                      "energy": (["time", "idx0"], np.tile(energy, (len(pad_arr), 1))),
                      "theta": (["time", "idx1"], np.tile(pitch_angles, (len(pad_arr), 1))),
                      "time": time,
                      "idx0": np.arange(len(energy)),
                      "idx1": np.arange(len(pitch_angles))})

    pad.attrs = vdf.attrs
    pad.attrs["mean_or_sum"] = mean_or_sum
    pad.attrs["delta_pitchangle_minus"] = d_angles * .5
    pad.attrs["delta_pitchangle_plus"] = d_angles * .5

    return pad



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""gse2gsm.py
@author: Louis Richard
"""

from .geocentric_coordinate_transformation import \
    geocentric_coordinate_transformation


def gse2gsm(inp, flag="gse>gsm"):
    """Converts GSE to GSM.

    Parameters
    ----------
    inp : xarray.DataArray or ndarray
        Time series of the input in GSE (GSM) coordinates.
        If ndarray first column is time in unix format.

    flag : str
        Flag for conversion direction. "gse>gsm" or "gsm>gse".
        Default is "gse>gsm"

    Returns
    -------
    out : xarray.DataArray or ndarray
        Time series of the input in GSM (GSE) coordinates.
        If ndarray first column is time in unix format.

    See also
    --------
    pyrfu.pyrf.geocentric_coordinate_transformation

    """

    assert flag in ["gse>gsm", "gsm>gse"], "invalid flag"

    out = geocentric_coordinate_transformation(inp, flag);

    return out


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2021 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""magnetopause_normal.py
@author: Louis Richard
"""

import warnings
import numpy as np

from scipy.optimize import fminbound


def magnetopause(theta, *args):
    """Magnetopause model to minimize.

    Parameters
    ----------
    theta : float
        to fill

    args : [float, float, float, float]
        extra arguments.

    Returns
    -------
    out : float
        to fill

    """

    r0, alpha, x0, y0 = args
    out = r0 ** 2 * (2. / (1 + np.cos(theta))) ** (2 * alpha) \
          - 2 * r0 * (2. / (1 + np.cos(theta))) ** alpha \
          * (x0 * np.cos(theta) + y0 * np.sin(theta)) + x0 ** 2 + y0 ** 2
    return out


def bow_shock(r_xy, *args):
    """Bow shock model to minimize.

    Parameters
    ----------
    r_xy : array_like
        to fill

    args : [float, float]
        extra arguments.

    Returns
    -------
    out : float
        to fill

    """
    x, y = r_xy
    x0, y0 = args
    out = (x - x0) ** 2 + (y - y0) ** 2
    return out


def magnetopause_normal(r_gsm, b_z_imf, p_sw, model: str = "mp_shue1997",
                        m_alfven: float = 4.):
    """Computes the distance and normal vector to the magnetopause for
    Shue et al., 1997 or Shue et al., 1998 model. Or bow shock for
    Farris & Russell 1994 model.

    Parameters
    ----------
    r_gsm : array_like
        GSM position in Re (if more than 3 values assumes that 1st is time).

    b_z_imf : float
        IMF Bz in nT.

    p_sw : float
        Solar wind dynamic pressure in nPa.

    model : str, optional
        Name of model :
            * 'mp_shue1997' : Shue et al., 1997 (Default)
            * 'mp_shue1998' : Shue et al., 1998
            * 'bs97'        : Bow shock, Farris & Russell 1994, based on
                                'mp_shue1997'
            * 'bs98'        : Bow shock, Farris & Russell 1994, based on
                                'mp_shue1998'

    m_alfven : float, optional
        Alfvenic Mach number, only needed if bow shock model is used.

    Returns
    -------
    mindist : float
        Minimum distance to the magnetopause, in Re. Positive value if
        spacecraft is inside the magnetopause, negative if outside the
        magnetopause.

    n_vec : ndarray
        Normal vector to the magnetopause (pointing away from Earth).


    References
    ----------
    .. [1]  J.-H. Shue, J. K. Chao, H. C. Fu, C. T. Russell, P. Song,
            K. K. Khurana, and H. J. Singer (1997), A new functional form to
            study the solar wind control of the magnetopause size and shape,
            J. Geophys. Res., 102, 9497, doi: https://doi.org/10.1029/97JA00196

    .. [2]  J.-H. Shue, P. Song, C. T. Russell, J. T. Steinberg, J. K. Chao,
            G. Zastenker, O. L. Vaisber, S. Kokubun, H. J. Singer,
            T. R. Detman and H. Kawano (1998), Magnetopause location under
            extreme solar wind conditions, J. Geophys. Res., 103(A8), 17,
            691–17,700, doi: https://doi.org/10.1029/98JA01103

    .. [3]  M. H. Farris and C. T. Russell (1994), Determining the standoff
            distance of the bow shock: Mach number dependence and use of
            models, J. Geophys. Res., 99, 17681.
            doi: https://doi.org/10.1029/94JA01020

    """

    if model in ["mp_shue1998", "bs98"]:
        alpha = (0.58 - 0.007 * b_z_imf) * (1.0 + 0.024 * np.log(p_sw))
        r0 = (10.22
              + 1.29 * np.tanh(.184 * (b_z_imf + 8.14))) * p_sw ** (-1. / 6.6)
        warnings.warn('Shue et al., 1998 model used.', UserWarning)
    else:
        alpha = (0.58 - 0.01 * b_z_imf) * (1.0 + 0.01 * p_sw)

        if b_z_imf >= 0:
            r0 = (11.4 + 0.013 * b_z_imf) * p_sw ** (-1.0 / 6.6)
        else:
            r0 = (11.4 + 0.140 * b_z_imf) * p_sw ** (-1.0 / 6.6)

        warnings.warn('Shue et al., 1997 model used.', UserWarning)

    # Spacecraft position
    r1_x, r1_y, r1_z = r_gsm
    r0_x, r0_y = [r1_x, np.sqrt(r1_y ** 2 + r1_z ** 2)]

    if model[:2].lower() == "mp":
        # Magnetopause

        theta_min, min_val, ierr, numfunc = fminbound(magnetopause,
                                                      x1=-np.pi / 2,
                                                      x2=np.pi / 2, args=(
            r0, alpha, r0_x, r1_y), full_output=True)

        min_dist = np.sqrt(min_val)

        # calculate the direction to the spacecraft normal to the magnetopause
        x_n = r0 * (2 / (1 + np.cos(theta_min))) ** alpha * np.cos(theta_min) \
              - r1_x
        phi = np.arctan2(r1_z, r1_y)
        y_n = np.cos(phi) * (r0 * (2 / (1 + np.cos(theta_min))) ** alpha *
                             np.sin(theta_min)) - r1_y
        z_n = np.sin(phi) * (r0 * (2 / (1 + np.cos(theta_min))) ** alpha
                             * np.sin(theta_min)) - r1_z

        n_vec = np.stack([x_n, y_n, z_n]) / min_dist

        # if statement to ensure normal is pointing away from Earth
        if (np.sqrt(r0_x ** 2 + r0_y ** 2)
                > r0 * (2 / (1 + np.cos(theta_min))) ** alpha):
            n_vec *= -1
            min_dist *= -1

    else:
        # Bow shock
        warnings.warn('Farris & Russell 1994 bow shock model used.',
                      UserWarning)

        gamma = 5 / 3
        mach = m_alfven

        # Bow shock standoff distance
        rbs = r0 * (1 + 1.1 * ((gamma - 1) * mach ** 2 + 2)
                    / ((gamma + 1) * (mach ** 2 - 1)))

        # y ^ 2 = 0 - Ax + Bx ^ 2
        coeffs = [0, 45.3, .04]
        x = np.linspace(rbs, -100, int((rbs + 100) * 1e3) + 1)
        y = np.sqrt(-coeffs[1] * (x - rbs) + coeffs[2] * (x - rbs) ** 2)
        x = np.hstack([np.flip(x), x])
        y = np.hstack([-np.flip(y), y])

        r_xy = np.transpose(np.stack([x, y]))
        args_bow_shock = (r0_x, r0_y)

        min_val = np.min(bow_shock(r_xy, *args_bow_shock))
        min_pos = np.argmin(bow_shock(r_xy, *args_bow_shock))

        d_vec = np.stack([x - r0_x, y - r0_y])
        d_min = d_vec[min_pos, :]

        x_n = d_min[0] / np.linalg.norm(d_min)
        min_dist = np.sqrt(min_val)

        qyz = r1_y / r1_z
        z_n = np.sign(r1_z) * np.sign(x_n) \
              * np.sqrt((1 - x_n ** 2) / (1 + qyz ** 2))
        y_n = z_n * qyz

        n_vec = np.stack([x_n, y_n, z_n])

        # if statement to ensure normal is pointing away from Earth
        if n_vec[0] < 0:
            n_vec *= -1
            min_dist *= -1

    return min_dist, n_vec



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

import numpy as np

from scipy import constants

from ..pyrf import resample, ts_scalar


def calculate_epsilon(vdf, model_vdf, n_s, sc_pot, **kwargs):
    """Calculates epsilon parameter using model distribution.

    Parameters
    ----------
    vdf : xarray.Dataset
        Observed particle distribution (skymap).

    model_vdf : xarray.Dataset
        Model particle distribution (skymap).

    n_s : xarray.DataArray
        Time series of the number density.

    sc_pot : xarray.DataArray
        Time series of the spacecraft potential.

    **kwargs : dict
        Hash table of keyword arguments with :
             * en_channels : list or numpy.ndarray
                Set energy channels to integrate over [min max]; min and max
                between must be between 1 and 32.

    Returns
    -------
    epsilon : xarray.DataArray
        Time series of the epsilon parameter.

    Examples
    --------
    >>> from pyrfu import mms
    >>> options = {"en_channel": [4, 32]}
    >>> eps = mms.calculate_epsilon(vdf, model_vdf, n_s, sc_pot, **options)

    """

    if np.abs(np.median(np.diff(vdf.time - n_s.time))).astype(float) > 0:
        raise RuntimeError("vdf and moments have different times.")

    # Default energy channels used to compute epsilon, lowest energy channel
    # should not be used.
    int_energies = np.arange(2, 33)

    if kwargs.get("en_channels"):
        assert isinstance(kwargs["en_channels"], (list, np.ndarray))
        int_energies = np.arange(kwargs["en_channels"][0],
                                 kwargs["en_channels"][1] + 1)

    # Resample sc_pot
    sc_pot = resample(sc_pot, n_s)

    # Remove zero count points from final calculation
    # model_vdf.data.data[vdf.data.data <= 0] = 0

    model_vdf /= 1e18
    vdf /= 1e18

    vdf_diff = np.abs(vdf.data.data - model_vdf.data.data)

    # Define constants
    q_e = constants.elementary_charge

    # Check whether particles are electrons or ions
    if vdf.attrs["specie"] == "e":
        m_s = constants.electron_mass
        print("notice : Particles are electrons")
    elif vdf.attrs["specie"] == "i":
        sc_pot.data *= -1
        m_s = constants.proton_mass
        print("notice : Particles are electrons")
    else:
        raise ValueError("Invalid specie")

    # Define lengths of variables
    n_ph = len(vdf.phi.data[0, :])

    # Define corrected energy levels using spacecraft potential
    energy_arr = vdf.energy.data
    v_s, delta_v = [np.zeros(energy_arr.shape) for _ in range(2)]

    for i in range(len(vdf.time)):
        energy_vec = energy_arr[i, :]
        energy_log = np.log10(energy_arr[i, :])

        v_s[i, :] = np.real(
            np.sqrt(2 * (energy_vec - sc_pot.data[i]) *q_e / m_s))

        temp0 = 2 * energy_log[0] - energy_log[1]
        temp33 = 2 * energy_log[-1] - energy_log[-2]

        energy_all = np.hstack([temp0, energy_log, temp33])

        diff_en_all = np.diff(energy_all)

        energy_upper = 10 ** (energy_log + diff_en_all[1:34] / 2)
        energy_lower = 10 ** (energy_log - diff_en_all[0:33] / 2)

        v_upper = np.sqrt(2 * q_e * energy_upper / m_s)
        v_lower = np.sqrt(2 * q_e * energy_lower / m_s)

        v_lower[np.isnan(v_lower)] = 0
        v_upper[np.isnan(v_upper)] = 0

        delta_v[i, :] = (v_upper - v_lower)

    v_s[v_s < 0] = 0

    # Calculate density of vdf_diff
    delta_ang = (11.25 * np.pi / 180) ** 2
    theta_k = vdf.theta.data

    epsilon = np.zeros(len(vdf.time))
    m_psd2_n = np.ones(n_ph) * np.sin(theta_k * np.pi / 180)

    for i in range(len(vdf.time)):
        for j in int_energies:
            tmp = np.squeeze(vdf_diff[i, j, ...])
            fct = v_s[i, j] ** 2 * delta_v[i, j] * delta_ang

            epsilon[i] += np.nansum(np.nansum(tmp * m_psd2_n, axis=0),
                                    axis=0) * fct

    epsilon /= 1e6 * n_s.data * 2
    epsilon = ts_scalar(vdf.time.data, epsilon)

    return epsilon



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""dsl2gse.py
©author: Louis Richard
"""

import warnings
import numpy as np
import xarray as xr

from astropy.time import Time

from ..pyrf import (geocentric_coordinate_transformation, resample,
                    sph2cart, ts_vec_xyz)


def dsl2gse(inp, defatt, direction=1):
    """Transform time series from DSL to GSE.

    Parameters
    ----------
    inp : xarray.DataArray
        Input time series to convert.

    defatt : xarray.Dataset or ndarray or list
        Spacecraft attitude.

    direction : int
        Direction of tranformation. +1 DSL -> GSE, -1 GSE -> DSL.
        Default is 1.

    Returns
    -------
    out : xarray.DataArray
        Time series of the input field in the new coordinates systems.


    Examples
    --------
    >>> from pyrfu.mms import get_data, load_ancillary, dsl2gse

    Define time interval

    >>> tint = ["2015-05-09T14:00:000", "2015-05-09T17:59:590"]

    Load magentic field in spacecraft coordinates

    >>> b_xyz = get_data("b_dmpa_fgm_brst_l2", tint, 1)

    Load spacecraft attitude

    >>> defatt = load_ancillary("defatt", tint, 1)

    Transform magnetic field to GSE

    >>> b_gse = dsl2gse(b_xyz, defatt)

    """

    if direction not in [-1, 1]:
        warnings.warn("using GSE->DSL", UserWarning)

    if isinstance(defatt, xr.Dataset):
        x, y, z = sph2cart(np.deg2rad(defatt.z_ra.data),
                           np.deg2rad(defatt.z_dec), 1)
        sax_gei = np.transpose(np.vstack([defatt.time.data.view("i8") * 1e-9,
                                          x, y, z]))
        sax_gse = geocentric_coordinate_transformation(sax_gei, "gei>gse")
        sax_gse = ts_vec_xyz(Time(sax_gse[:, 0], format="unix").datetime64,
                             sax_gse[:, 1:])

        spin_ax_gse = resample(sax_gse, inp)
        spin_axis = spin_ax_gse.data

    elif isinstance(defatt, (np.ndarray, list)) and len(defatt) == 3:
        spin_axis = defatt

    else:
        raise ValueError("unrecognized DEFATT/SAX input")

    r_x, r_y, r_z = [spin_axis[:, i] for i in range(3)]

    a = 1. / np.sqrt(r_y ** 2 + r_z ** 2)
    transf_mat = np.zeros((len(a), 3, 3))
    transf_mat[:, 0, :] = np.transpose(np.stack([a * (r_y ** 2 + r_z ** 2),
                                                 -a * r_x * r_y,
                                                 -a * r_x * r_z]))

    transf_mat[:, 1, :] = np.transpose(np.stack([0. * a, a * r_z, -a * r_y]))

    transf_mat[:, 2, :] = np.transpose(np.stack([r_x, r_y, r_z]))

    if direction == 1:
        transf_mat = np.transpose(transf_mat, [0, 2, 1])

    out_data = np.einsum('kji,ki->kj', transf_mat, inp.data)

    out = inp.copy()
    out.data = out_data

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""__init__.py
@author: Louis Richard
"""

# @Louis Richard
from .split_vs import split_vs
from .list_files import list_files
from .get_ts import get_ts
from .get_dist import get_dist
from .get_data import get_data
from .db_get_ts import db_get_ts
from .fk_power_spectrum_4sc import fk_power_spectrum_4sc
from .lh_wave_analysis import lh_wave_analysis
from .whistler_b2e import whistler_b2e
from .get_feeps_energy_table import get_feeps_energy_table
from .get_feeps_active_eyes import get_feeps_active_eyes
from .get_feeps_oneeye import get_feeps_oneeye
from .get_feeps_omni import get_feeps_omni
from .get_feeps_alleyes import get_feeps_alleyes
from .read_feeps_sector_masks_csv import read_feeps_sector_masks_csv
from .feeps_split_integral_ch import feeps_split_integral_ch
from .feeps_remove_sun import feeps_remove_sun
from .calc_feeps_omni import calc_feeps_omni
from .feeps_spin_avg import feeps_spin_avg
from .feeps_pitch_angles import feeps_pitch_angles
from .calc_feeps_pad import calc_feeps_pad
from .get_eis_allt import get_eis_allt
from .get_eis_omni import get_eis_omni
from .remove_idist_background import remove_idist_background
from .psd_moments import psd_moments
from .rotate_tensor import rotate_tensor
from .calculate_epsilon import calculate_epsilon
from .dft_time_shift import dft_time_shift
from .estimate_phase_speed import estimate_phase_speed
from .fft_bandpass import fft_bandpass
from .get_pitch_angle_dist import get_pitch_angle_dist
from .make_model_vdf import make_model_vdf
from .psd_rebin import psd_rebin
from .load_ancillary import load_ancillary
from .vdf_omni import vdf_omni
from .spectr_to_dataset import spectr_to_dataset
from .vdf_to_deflux import vdf_to_deflux
from .vdf_to_dpflux import vdf_to_dpflux
from .dpflux_to_vdf import dpflux_to_vdf
from .deflux_to_vdf import deflux_to_vdf
from .copy_files import copy_files
from .vdf_to_e64 import vdf_to_e64
from .current_location import current_location
from .dsl2gse import dsl2gse



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""ts_time.py
@author: Louis Richard
"""

import numpy as np
import xarray as xr

from astropy.time import Time


def ts_time(time, fmt="unix"):
    """Creates time line in DataArray.

    Parameters
    ----------
    time : ndarray
        Input time line.

    fmt : str
        Format of the input time line.

    Returns
    -------
    out : xarray.DataArray
        Time series of the time line.

    """

    assert isinstance(time, np.ndarray)

    time = Time(time, format=fmt).datetime64

    out = xr.DataArray(time, coords=[time], dims=["time"])

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""e_vxb.py
@author: Louis Richard
"""

import numpy as np

from .resample import resample
from .ts_vec_xyz import ts_vec_xyz


def e_vxb(v_xyz, b_xyz, flag="vxb"):
    """Computes the convection electric field :math:`\\mathbf{V}\\times
    \\mathbf{B}` (default) or the :math:`\\mathbf{E}\\times\\mathbf{
    B}/|\\mathbf{B}|^{2}` drift velocity (flag="exb").

    Parameters
    ----------
    v_xyz : xarray.DataArray
        Time series of the velocity/electric field.

    b_xyz : xarray.DataArray
        Time series of the magnetic field.

    flag : str
        Method flag :
            * "vxb" : computes convection electric field (default).
            * "exb" : computes ExB drift velocity.

    Returns
    -------
    out : xarray.DataArray
        Time series of the convection electric field/ExB drift velocity.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic field and electric field

    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_fast_l2", tint, mms_id)

    Compute ExB drift velocity

    >>> v_xyz_exb = pyrf.e_vxb(e_xyz, b_xyz,"ExB")

    """

    estimate_exb = False
    input_v_cons = False

    if flag.lower() == "exb":
        estimate_exb = True

    if v_xyz.size == 3:
        input_v_cons = True

    if estimate_exb:
        if len(v_xyz) != len(b_xyz):
            b_xyz = resample(b_xyz, v_xyz)

        res = 1e3 * np.cross(v_xyz.data, b_xyz.data, axis=1)
        res /= np.linalg.norm(b_xyz.data, axis=1)[:, None] ** 2

        attrs = {"UNITS": "km/s", "FIELDNAM": "Velocity",
                 "LABLAXIS": "V"}

    else:
        if input_v_cons:
            res = np.cross(np.tile(v_xyz, (len(b_xyz), 1)), b_xyz.data) * (-1) * 1e-3

        else:
            if len(v_xyz) != len(b_xyz):
                b_xyz = resample(b_xyz, v_xyz)

            res = np.cross(v_xyz.data, b_xyz.data) * (-1) * 1e-3

        attrs = {"UNITS": "mV/s", "FIELDNAM": "Electric field",
                 "LABLAXIS": "E"}

    out = ts_vec_xyz(b_xyz.time.data, res, attrs)

    return out
